library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity RAM_DAC is
    Port (Palette : in  STD_LOGIC_VECTOR (3 downto 0);  --Palette P[3:0] inputs. P4 is always high and P[7:6] are always low in the scope.
          Overlay : in  STD_LOGIC_VECTOR (3 downto 0);  -- Overlay O[3:0] inputs
		  Data : in STD_LOGIC_VECTOR (7 downto 0);  -- Data inputs D[7:0]
          RS : in  STD_LOGIC_VECTOR (2 downto 0);  --Register Select inputs R[2:0]
          WR_bar: in STD_LOGIC;  --/WR input
          clk : in  STD_LOGIC;  -- 25MHz clock input
		  blank_bar : in STD_LOGIC;  --/blank input
          LCD_Out1_N, LCD_Out2_N, LCD_Out3_N : out STD_LOGIC;  -- lvds outputs to drive LCD inputs
          LCD_Out1_P, LCD_Out2_P, LCD_Out3_P : out STD_LOGIC;
          LCD_Clock_Out_N : out STD_LOGIC;  --  LVDS clock to LCD
          LCD_Clock_Out_P : out STD_LOGIC;
          BL : in STD_LOGIC_VECTOR (2 downto 0);  --  inputs to set lcd backlight brightness (PWM).  From DIP switch on board. 000 = dimmest, 111 = brightest
          PWM : out STD_LOGIC;  -- PWM signal for LCD backlight.  Output to lcd to control backlight level
          LED_ON : out STD_LOGIC; -- LCD Backlight on/off control.  Turns on lcd backlight.  Always 1.
          --DEBUG
          led1 : buffer STD_LOGIC  --PLL locked indication
          --test_port_1 : out std_logic;
          --Debug_out : out std_logic_vector (3 downto 0)
	);
		
end RAM_DAC;

architecture Behavioral of RAM_DAC is
--  Components
    component clk_wiz_0  --Auto generated with the Vivado clock wizard.  Creates 175MHz, 25MHz and 5MHz clocks.  
        port (
          clk_out175M          : out    std_logic;  --for clocking the lcd lvds serial outputs
          clk_out25M          : out    std_logic;  --for clocking internal logic.  Allows phase adjust from incoming 25MHz clock if necessary.
          clk_out5M          : out    std_logic;  --for clocking the pwm register for the lcd backlight pwm.
          locked            : out    std_logic;  --visual indication of lock for the clocks
          clk_in25           : in     std_logic  --25MHz clock input
     );
    end component;
    
    component OBUFDS 
        port(
            I : in std_logic; 
            O : out std_logic; 
            OB : out std_logic); 
        end component; 
    
 signal clk_175 : std_logic;  --175MHz clock from clock generator.  Phase locked to input 25MHz pixel clock
 signal clk_25 : std_logic;  --25MHz clock from clock generator.  Phase locked to input 25MHz pixel clock.  Phase can be adjusted it required for timing if requied (using clock wizard)
 signal clk_5 : std_logic;  --5MHz clock from clock generator.  Phase locked to input 25MHz pixel clock.  Used by backlight led pwm.
 signal locked : std_logic;  --locked fron clock generator
 signal register_mode : std_logic_vector (2 downto 0);  --Set by RS[2:0].  Latched on falling edge of WR_bar
 signal palette_address_register : std_logic_vector (7 downto 0);  --256 pallet RAM locations
 signal overlay_address_register : std_logic_vector (7 downto 0);  --15 overlay RAM locations
 type array_16x6 is array(15 downto 0) of std_logic_vector (5 downto 0);
 signal palette_RAM_red: array_16x6;  --only 16 locations are used in the scope.
 signal palette_RAM_green: array_16x6;
 signal palette_RAM_blue: array_16x6;
 signal overlay_RAM_red: array_16x6;  --There are only 15 overlay registers because location 0000 is reserved.
 signal overlay_RAM_green: array_16x6;
 signal overlay_RAM_blue: array_16x6;
 signal lvds_count: natural range 0 to 7; --used to count serial output clocks for the lcd lvds outputs
 signal synced: std_logic;  --used to sync the lvds_count with the start of the 25MHz clock
 signal de : std_logic;  --latched version of blank_bar.  Latched on rising edge of 25MHz clock
 signal red, grn, blu : std_logic_vector(5 downto 0);  --Latched version of the outputs of the palette and overlay RAM.  Latched on rising edge of 25MHz clock
 signal LCD_Out1, LCD_Out2, LCD_Out3 : STD_LOGIC;  -- outputs to lvds converters to drive LCD inputs
 signal LCD_Clock_Out : STD_LOGIC;  --  LVDS clock to LCD
 
attribute IOSTANDARD : string; 
attribute IOSTANDARD of lvds_out_1 : label is "LVDS_25"; 
attribute IOSTANDARD of lvds_out_2 : label is "LVDS_25";
attribute IOSTANDARD of lvds_out_3 : label is "LVDS_25"; 
attribute IOSTANDARD of lvds_out_clk : label is "LVDS_25"; 
 
begin
-- Connect Components
    clock_generator : clk_wiz_0
       port map ( 
      -- Clock out ports  
       clk_out175M => clk_175,
       clk_out25M => clk_25,
       clk_out5M => clk_5,
      -- Status and control signals                1
       locked => locked,
       -- Clock in ports
       clk_in25 => clk
     );
     
     lvds_out_1 :  OBUFDS -- LVDS output buffer
        port map (
            I => LCD_Out1, 
            O => LCD_Out1_P, 
            OB => LCD_Out1_N); 
            
     lvds_out_2 :  OBUFDS -- LVDS output buffer
        port map (
            I => LCD_Out2, 
            O => LCD_Out2_P, 
            OB => LCD_Out2_N); 
            
      lvds_out_3 :  OBUFDS -- LVDS output buffer
        port map (
            I => LCD_Out3, 
            O => LCD_Out3_P, 
            OB => LCD_Out3_N); 
            
       lvds_out_clk :  OBUFDS -- LVDS output buffer
        port map (
            I => LCD_Clock_Out, 
            O => LCD_Clock_Out_P, 
            OB => LCD_Clock_Out_N); 
     
     
 store_register_mode: process (WR_bar) -- store the register mode at the falling edge of WR_bar
 begin
    if falling_edge(WR_bar) then  --on falling edge of WR_bar, save the register mode and latch the address
        register_mode <= RS; 
     end if;
 end process store_register_mode;
     
 write_RAM: process (WR_bar)  --write the palette and overlay RAM on the rising edge of WR_bar.  
    variable address_a_b : natural range 0 to 2 := 0; --red/green/blue address counter bits(address_a address_b in data sheet)
 begin
    if rising_edge(WR_bar) then  --on rising edge of WR_bar, store the data in the address or RAM
            case (register_mode) is
                when "000" =>  -- Pallet RAM address register write (RAM write mode)
                    palette_address_register <= Data;
                    address_a_b := 0;
                when "100" =>  -- Overlay RAM address register write (RAM write mode)
                    overlay_address_register <= Data;
                    address_a_b := 0;
                when "001" =>  --Pallet RAM write
                    if address_a_b = 0 then
                        palette_RAM_red(to_integer(unsigned(palette_address_register(3 downto 0)))) <= Data(5 downto 0); --write to red pallet RAM
                    elsif address_a_b = 1 then
                        palette_RAM_green(to_integer(unsigned(palette_address_register(3 downto 0)))) <= Data(5 downto 0); --write to green pallet RAM
                    elsif address_a_b = 2 then
                        palette_RAM_blue(to_integer(unsigned(palette_address_register(3 downto 0)))) <= Data(5 downto 0); --write to blue pallet RAM
                    end if;
                    if address_a_b < 2 then
                        address_a_b := address_a_b + 1;
                    else
                        address_a_b := 0;
                        palette_address_register <= std_logic_vector(unsigned(palette_address_register) + 1);
                    end if;
                when "101" =>  --Overlay RAM write
                    if address_a_b = 0 then
                        overlay_RAM_red(to_integer(unsigned(overlay_address_register(3 downto 0)))) <= Data(5 downto 0); --write to red pallet RAM
                    elsif address_a_b = 1 then
                        overlay_RAM_green(to_integer(unsigned(overlay_address_register(3 downto 0)))) <= Data(5 downto 0); --write to green pallet RAM
                    elsif address_a_b = 2 then
                        overlay_RAM_blue(to_integer(unsigned(overlay_address_register(3 downto 0)))) <= Data(5 downto 0); --write to blue pallet RAM
                    end if;
                    if address_a_b < 2 then
                        address_a_b := address_a_b + 1;
                    else
                        address_a_b := 0;
                        overlay_address_register <= std_logic_vector(unsigned(overlay_address_register) + 1);
                    end if;
                when others =>  --other functions (pixel mask, register reads) are not used by the scope.
                    address_a_b := 0;
            end case;            
  end if;
 end process write_RAM; 
     

load_output_data : process(clk_25)  --latch the RAM outputs on the rising edge of the 25MHz clock (start of the pixel time)
begin
if rising_edge(clk_25) then
    de <= blank_bar;
    case (Overlay) is
        when "0000" =>
            red <= palette_RAM_red(to_integer(unsigned(palette)));
            grn <= palette_RAM_green(to_integer(unsigned(palette)));
            blu <= palette_RAM_blue(to_integer(unsigned(palette)));
        when others =>
            red <= overlay_RAM_red(to_integer(unsigned(overlay)));
            grn <= overlay_RAM_green(to_integer(unsigned(overlay)));
            blu <= overlay_RAM_blue(to_integer(unsigned(overlay)));
    end case;
end if;
end process load_output_data;

synchronize_lvds : process(clk_25)  --force the lvds_count to start at the beginning of the 25MHz pixel clock 
begin
    if rising_edge(clk_25) then
        if lvds_count = 0 then
            synced <= '1';
        else
            synced <= '0';
        end if;
    end if;
end process synchronize_lvds;

lvds_counter : process(clk_175)  --count the lvds output position to but the correct signal on the lvds serial outputs.
begin
if falling_edge(clk_175) then
    if (synced = '0') then
        lvds_count <= 0;
    elsif lvds_count = 6 then
        lvds_count <= 0;
    else
        lvds_count <= lvds_count + 1; 
    end if;
end if;
end process lvds_counter;
      
update_pixel : process(clk_175) --send out the lvds signals to the lcd
begin
if falling_edge(clk_175) then
case (lvds_count) is
    when 0 =>
        LCD_Out1 <= grn(0);
        LCD_Out2 <= blu(1);
        LCD_Out3 <= de;
        LCD_Clock_Out <= '1';
   when 1 =>
        LCD_Out1 <= red(5);
        LCD_Out2 <= blu(0);
        LCD_Out3 <= '0';
        LCD_Clock_Out <= '1';
  when 2 =>
        LCD_Out1 <= red(4);
        LCD_Out2 <= grn(5);
        LCD_Out3 <= '0';
        LCD_Clock_Out <= '0';
  when 3 =>
        LCD_Out1 <= red(3);
        LCD_Out2 <= grn(4);
        LCD_Out3 <= blu(5);
        LCD_Clock_Out <= '0';
  when 4 =>
        LCD_Out1 <= red(2);
        LCD_Out2 <= grn(3);
        LCD_Out3 <= blu(4);
        LCD_Clock_Out <= '0';
  when 5 =>
        LCD_Out1 <= red(1);
        LCD_Out2 <= grn(2);
        LCD_Out3 <= blu(3);
        LCD_Clock_Out <= '1';
  when 6 =>
        LCD_Out1 <= red(0);
        LCD_Out2 <= grn(1);
        LCD_Out3 <= blu(2);
        LCD_Clock_Out <= '1';
  when others =>   
        LCD_Out1 <= red(0);
        LCD_Out2 <= grn(1);
        LCD_Out3 <= blu(2);
        LCD_Clock_Out <= '1';
 end case;   
end if;
end process update_pixel;

lcd_backlight_pwm : process (clk_5)  --set the lcd backlight brightness (pwm) based on the DIP switch setting
    variable pwm_count: unsigned(7 downto 0) := "00000000";
    variable brightness_value: unsigned(7 downto 0) := "00000000";
    begin
    led_on <= '1';
    if rising_edge(clk_5) then
        brightness_value(7 downto 5) := unsigned(BL);
        pwm_count := pwm_count + 1;
        if pwm_count = 255 then
            pwm_count := "00000000";
        end if;
        if to_integer(unsigned(brightness_value)) >  to_integer(pwm_count) then
            pwm <= '1';
        else
            pwm <= '0';
        end if; 
    end if;
end process lcd_backlight_pwm;

led1 <= locked;  --visual indication of the lock signal from the PLL
	
end Behavioral;



